#!/usr/bin/env python3

import argparse
import logging

import signal
import time
from multiprocessing import Pool as ThreadPool
import os

from lofar.calibration.common.utils import *
from lofar.calibration.common.datacontainers.holography_dataset import HolographyDataset

DEFAULT_SLEEP_TIME = 1

logger = logging.getLogger('mshologextract')


def main():
    cla_parser = specify_command_line_arguments()
    arguments = parse_command_line_arguments_and_set_verbose_logging(cla_parser)
    read_holography_datasets_and_store(arguments.input_path, arguments.output_path,
                                       holography_bsf_path=arguments.holography_bsf,
                                       holography_ms_path=arguments.holography_ms,
                                       n_processes=arguments.num_proc)


def parse_command_line_arguments_and_set_verbose_logging(parser):
    """
    Given a parser it parses the command line arguments and return the result of the parsing.

    :param parser: command line parser
    :type parser: argparse.ArgumentParser
    :return: the parsed arguments
    :rtype: argparse.Namespace
    """
    arguments=parser.parse_args()
    if arguments.v:
        logging.basicConfig(logging=logging.DEBUG)
    else:
        logging.basicConfig(logging=logging.INFO)
    return arguments


def specify_command_line_arguments():
    """
    Setup the command argument parser and returns it
    :return: the parser for the command line arguments
    :rtype: argparse.ArgumentParser
    """
    parser = argparse.ArgumentParser(description='This program is meant for convert the Holography'
                                                 ' observation\'s data into an holography dataset')
    parser.add_argument('input_path', help='path to the holography observation data')
    parser.add_argument('output_path', help='path to the holography observation data')

    parser.add_argument('--holography_bsf', help='override default path for the holography beam'
                                                 ' specification file', default=None)
    parser.add_argument('--holography_ms', help='override default path for the holography'
                                                ' observation MS files', default=None)
    parser.add_argument('--num_proc', help='number of processes used to convert the holography'
                                           ' observation', type=int)
    parser.add_argument('-v', help='verbose logging', action='store_true')

    return parser


def read_and_store_single_target_station(target_station_name, matched_bsf_ms_pair, store_path):
    """
    Reads a single target station hdf5 file from a list of bsf and ms pairs
    :param target_station_name: target station name to process
    :type target_station_name: str
    :param matched_bsf_ms_pair: the list containing a tuple with the holography specification and
    its specific holography observation
    :type matched_bsf_ms_pair: list[(lofar.calibration.common.datacontainers.HolographySpecification,
                                    lofar.calibration.common.datacontainers.HolographyObservation)]
    :param store_path: path where to store the file
    :type store_path: str
    :return:
    """
    holography_dataset = HolographyDataset()
    logger.info('Loading data for station %s ', target_station_name)
    holography_dataset.load_from_beam_specification_and_ms(target_station_name,
                                                           matched_bsf_ms_pair)
    filename = '%s.hdf5' % target_station_name
    outpath = os.path.join(store_path, filename)

    logger.info('Storing data for station %s in %s', target_station_name, outpath)

    holography_dataset.store_to_file(outpath)

    logger.info('Stored data for station %s in %s', target_station_name, outpath)


def make_worker_ignore_siginterrupt():
    """
    Force the worker process to ignore keyboard interrupt signals
    """
    signal.signal(signal.SIGINT, signal.SIG_IGN)


def wait_for_processes_to_complete_execution(thread_pool,
                                             future_results,
                                             sleep_time):
    """
    Wait for the processess to complete execution if KeyboardInterrupt signal is sent
    it terminates all the processes and wait for them to terminated.
    :param thread_pool: pool of processes
    :type thread_pool: multiprocess.Pool
    :param future_results: list of result objects
    :type future_results: list[multiprocess.pool.ApplyResult]
    """
    all_done = False
    print(

        'dude'
    )
    try:
        while not all_done:
            all_done = True
            for k in future_results:
                all_done = all_done and k.ready()
                print(k)
                print(future_results)
            time.sleep(sleep_time)
    except KeyboardInterrupt:
        thread_pool.terminate()
        thread_pool.join()

    thread_pool.close()
    thread_pool.join()


def read_holography_datasets_and_store(holography_observation_path,
                                       store_path,
                                       holography_bsf_path=None,
                                       holography_ms_path=None,
                                       n_processes=1,
                                       sleep_time=DEFAULT_SLEEP_TIME):

    """
    Read the holography datasets and store them in a the specified directory
    :param holography_observation_path: path to the holography observation
    :type holography_observation_path: str
    :param store_path: path were
    :type store_path: str
    :param holography_bsf_path: path to the beam specification file
    :type holography_bsf_path: str
    :param holography_ms_path: path to the measurement set directory
    :type holography_ms_path: str
    :param n_processes: number of precesses to spawn to do the conversion
    :type n_processes: int
    """

    if holography_ms_path is not None:
        raise NotImplementedError()
    if holography_bsf_path is not None:
        raise NotImplementedError()
    target_station_names = list_all_target_stations(holography_observation_path)

    logger.debug('target station names %s', target_station_names)
    matched_bsf_ms_pair =\
        match_holography_beam_specification_file_with_observation(holography_observation_path)
    logger.debug('matched beam specification files and measurement sets: %s',
                 matched_bsf_ms_pair)
    try:
        if not os.path.exists(store_path):
            os.makedirs(store_path)
    except Exception as e:
        logger.exception('Cannot create output path %s: %s', store_path, e)
        raise e

    thread_pool = ThreadPool(n_processes, make_worker_ignore_siginterrupt)
    results = []
    for station_name in target_station_names:
        result = thread_pool.apply_async(read_and_store_single_target_station,
                               (station_name, matched_bsf_ms_pair, store_path))

        results.append(result)

    wait_for_processes_to_complete_execution(thread_pool, results, sleep_time)


if __name__ == '__main__':
    logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.DEBUG)
    main()