#!/usr/bin/env python3
import argparse
import sys
import os
import lofar.calibration.common.datacontainers as datacontainers
from lofar.calibration.processing.solver import solve_gains_per_datatable
from lofar.calibration.processing.averaging import average_data,\
    weighted_average_dataset_per_station
from lofar.calibration.processing.interpolate import derive_interpolation_parameters
from lofar.calibration.processing.normalize import normalize_beams_by_central_one
import logging

import functools

logger = logging.getLogger('holography_process')


def log_step_execution(step_description):
    def log_step_decorator(function):
        @functools.wraps(function)
        def wrapper(*args, **kwargs):
            logger.info('Performing step %s', step_description)
            try:
                result = function(*args, **kwargs)
                logger.info('Step %s performed', step_description)
                return result

            except Exception as e:
                logger.exception('exception occurred performing step %s: %s',
                                 step_description, e)
                raise e

        return wrapper
    return log_step_decorator


def setup_argument_parser():
    """

    :return:
    """
    parser = argparse.ArgumentParser(
        description='Computes the gains per each antenna from the given Holography Dataset')
    parser.add_argument('input_path', help='Input holography dataset', type=str)
    parser.add_argument('output_path', help='Output holography dataset', type=str)
    parser.add_argument('verbose', help='Set verbose logging', action='store_false')
    return parser


def parse_command_arguments(arguments):
    """

    :param arguments:
    :type arguments: list
    :return:
    """
    parser = setup_argument_parser()
    return parser.parse_args(arguments)


@log_step_execution('load')
def loading(path):
    """

    :param path:
    :type path: str
    :return:
    :rtype: datacontainers.HolographyDataset
    """
    dataset = datacontainers.HolographyDataset.load_from_file(path)
    return dataset


@log_step_execution('normalize')
def normalize_step(dataset, input_datatable):
    """

    :param dataset:
    :type dataset: datacontainers.HolographyDataset
    :param input_datatable: dict(dict(dict(numpy.ndarray)))
    :return:
    :rtype: dict(dict(dict(numpy.ndarray)))
    """
    output_datatable = normalize_beams_by_central_one(dataset,
                                                      input_datatable,
                                                      dataset.central_beamlets)
    dataset.derived_data['NORMALIZED'] = output_datatable
    return output_datatable


@log_step_execution('time average')
def time_averaging_step(dataset, input_datatable):
    """

    :param dataset:
    :type dataset: datacontainers.HolographyDataset
    :param input_datatable: dict(dict(dict(numpy.ndarray)))
    :return:
    :rtype: dict(dict(dict(numpy.ndarray)))
    """
    output_datable = average_data(input_datatable)
    dataset.derived_data['TIME_AVERAGED'] = output_datable
    return output_datable


@log_step_execution('station average')
def station_averaging_step(dataset, input_datatable):
    """

    :param dataset:
    :type dataset: datacontainers.HolographyDataset
    :param input_datatable: dict(dict(dict(numpy.ndarray)))
    :return:
    :rtype: dict(dict(dict(numpy.ndarray)))
    """
    output_datable = weighted_average_dataset_per_station(dataset, input_datatable)
    if dataset.derived_data is None:
        dataset.derived_data = dict()
    dataset.derived_data['STATION_AVERAGED'] = output_datable
    return output_datable


@log_step_execution('solving gains')
def compute_gains_step(dataset, input_datatable, direct_complex=True):
    """

    :param dataset:
    :type dataset: datacontainers.HolographyDataset
    :param input_datatable: dict(dict(dict(numpy.ndarray)))
    :return:
    :rtype: dict(dict(dict(numpy.ndarray)))
    """
    output_datable = solve_gains_per_datatable(dataset, input_datatable, direct_complex=direct_complex)
    if dataset.derived_data is None:
        dataset.derived_data = dict()
    dataset.derived_data['GAINS'] = output_datable
    return output_datable

@log_step_execution('gains interpolation')
def interpolate_gains_step(dataset, input_datatable):
    """

    :param dataset:
    :type dataset: datacontainers.HolographyDataset
    :param input_datatable: dict(dict(dict(numpy.ndarray)))
    :return:
    :rtype: dict(dict(dict(numpy.ndarray)))
    """
    output_datable = derive_interpolation_parameters(input_datatable)
    if dataset.derived_data is None:
        dataset.derived_data = dict()
    dataset.derived_data['INTERPOLATED_GAINS'] = output_datable
    return output_datable

@log_step_execution('save')
def store_step(dataset, filepath):
    absolute_filepath = os.path.abspath(filepath)
    dataset.store_to_file(absolute_filepath)


def prepare_dataset_for_processing(dataset: datacontainers.HolographyDataset):
    """
    :param dataset:
    :return: None
    """
    if dataset.derived_data is None:
        dataset.derived_data = dict()

def execute_processing(arguments):
    dataset = loading(arguments.input_path)
    prepare_dataset_for_processing(dataset)
    normalized_data = normalize_step(dataset, dataset.data)
    averaged_data = time_averaging_step(dataset, normalized_data)
    station_averaged_data = station_averaging_step(dataset, averaged_data)

    gains = compute_gains_step(dataset, station_averaged_data)
    interpolate_gains_step(dataset, gains)

    store_step(dataset, arguments.output_path)


def setup_logger(arguments):
    logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    if arguments.verbose:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)


if __name__ == '__main__':
    arguments = parse_command_arguments(sys.argv[1:])
    setup_logger(arguments)
    execute_processing(arguments)


