#!/usr/bin/env python3
import argparse
import sys
from lofar.calibration.processing.inspect import plot_station_averaged_visibilities_lm_plane, \
                                                 plot_gains_per_antenna, plot_gains_fit_per_antenna



import matplotlib as mpl

import lofar.calibration.common.datacontainers as datacontainers

import logging

import functools

logger = logging.getLogger('holography_process')


def setup_matplotlib():
    if 'cairo' in mpl.rcsetup.non_interactive_bk:
        mpl.rcParams['backend'] = 'cairo'
    elif 'agg' in mpl.rcsetup.non_interactive_bk:
        mpl.rcParams['backend'] = 'agg'
    else:
        raise ModuleNotFoundError('Please install either cairo or agg backend')

    mpl.rcParams['figure.autolayout'] = True


def log_step_execution(step_description):
    def log_step_decorator(function):
        @functools.wraps(function)
        def wrapper(*args, **kwargs):
            logger.info('Performing step %s', step_description)
            try:
                result = function(*args, **kwargs)
                logger.info('Step %s performed', step_description)
                return result

            except Exception as e:
                logger.exception('exception occurred performing step %s: %s',
                                 step_description, e)
                raise e

        return wrapper
    return log_step_decorator


def setup_argument_parser():
    """

    :return:
    """
    parser = argparse.ArgumentParser(
        description='Inspect the processed holography dataset')
    parser.add_argument('input_path', help='Input holography dataset', type=str)
    parser.add_argument('output_path', help='Output holography dataset', type=str)
    parser.add_argument('verbose', help='Set verbose logging', action='store_false')
    return parser


def parse_command_arguments(arguments):
    """

    :param arguments:
    :type arguments: list
    :return:
    """
    parser = setup_argument_parser()
    return parser.parse_args(arguments)


@log_step_execution('load')
def loading(path):
    """

    :param path:
    :type path: str
    :return:
    :rtype: datacontainers.HolographyDataset
    """
    dataset, f = datacontainers.HolographyDataset.open_holography_file(path)
    return dataset, f


@log_step_execution('plot station averaged visibilities')
def plot_averaged_data(dataset: datacontainers.HolographyDataset, save_to):
    try:
        plot_station_averaged_visibilities_lm_plane(dataset, save_to=save_to)
    except KeyError:
        logger.error('Station averaged visibilities not present in the holography'
                     ' dataset. Skipping plot generation...')


@log_step_execution('plot gains per frequency')
def plot_gains(dataset: datacontainers.HolographyDataset, save_to):
    try:
        plot_gains_per_antenna(dataset, save_to=save_to)
    except KeyError:
        logger.error('Gains not present in the holography dataset.'
                     ' Skipping plot generation...')


@log_step_execution('plot gains fit per frequency')
def plot_gains_fit(dataset: datacontainers.HolographyDataset, save_to):
    try:
        plot_gains_fit_per_antenna(dataset, save_to=save_to)
    except KeyError:
        logger.error('Gains fit not present in the holography dataset. Skipping plot generation...')


def close_file(f):
    if f is not None:
        f.close()


def execute_processing(arguments):
    dataset, file_descriptor = loading(arguments.input_path)
    plot_averaged_data(dataset, save_to=arguments.output_path)
    plot_gains(dataset, save_to=arguments.output_path)
    plot_gains_fit(dataset, save_to=arguments.output_path)
    close_file(file_descriptor)


def setup_logger(arguments):
    logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    if arguments.verbose:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)


if __name__ == '__main__':
    arguments = parse_command_arguments(sys.argv[1:])
    setup_logger(arguments)
    setup_matplotlib()
    execute_processing(arguments)
