import typing

import matplotlib.pyplot as plt
import os
from lofar.calibration.common.datacontainers.holography_dataset import HolographyDataset
from matplotlib import cm
from matplotlib.figure import Figure

from scipy.constants import c as light_speed
from numpy.fft import fft2, fftshift

import numba
from numba import jit
import numpy

from matplotlib.ticker import AutoMinorLocator


def __set_minor_ticks_for_canvas(canvas):
    canvas.xaxis.set_minor_locator(AutoMinorLocator())
    canvas.yaxis.set_minor_locator(AutoMinorLocator())


__LIST_OR_NPARRAY = typing.Union[typing.List[float], numpy.ndarray]
__DATATABLE_TYPE = typing.Dict[
    str, typing.Dict[str, typing.Dict[str, typing.Dict[str, numpy.ndarray]]]]
__MHZ_IN_HZ = 1.e6

DEFAULT_FORMAT = 'png'


def __save_figure_to_path(figure, path, title, format=DEFAULT_FORMAT):
    os.makedirs(path, exist_ok=True)
    full_path = os.path.join(path, title)
    figure.savefig(full_path, format=format)


@jit()
def dft2_numba(x: numpy.ndarray,
               y: numpy.ndarray,
               l: numpy.ndarray,
               m: numpy.ndarray,
               freq_hz,
               v: numpy.ndarray,
               fourier_sign):

    arg_factor = (fourier_sign * 2 * numpy.pi * freq_hz / light_speed)
    result = numpy.zeros((len(x), len(y)), dtype=numpy.complex64)
    x_down = x.astype(numpy.float32)
    y_down = y.astype(numpy.float32)
    l_down = l.astype(numpy.float32)
    m_down = m.astype(numpy.float32)
    v_down = v.astype(numpy.complex64)

    dft2_numba_loop(x_down, y_down, l_down, m_down, v_down, arg_factor, result)
    return result


@jit(parallel=True, fastmath=True)
def dft2_numba_loop(x, y, l, m, v, arg_factor, result):
    y_len = len(y)
    x_len = len(x)
    k_len = len(v)
    phases = numpy.zeros((x_len, y_len, k_len))
    for i in numba.prange(y_len):
        for j in range(x_len):
            for k in range(v.shape[0]):
                phases[i, j, k] = arg_factor * (x[j] * l[k] + y[i] * m[k])

    real_factor = numpy.cos(phases)
    imag_factor = numpy.cos(phases)

    for i in numba.prange(y_len):
        for j in range(x_len):
            for k in range(v.shape[0]):

                result[i, j] += v[k] * (real_factor[i, j, k] + 1.j * imag_factor[i, j, k])


def complex_value_to_color(complex_array: numpy.ndarray, abs_max=None, abs_min=None, log=True):
    linearized_complex_array = complex_array.ravel()

    abs_array = numpy.abs(linearized_complex_array)

    if abs_max is None:
        abs_max = numpy.max(abs_array)

    if abs_min is None:
        abs_min = numpy.min(abs_array)
    if log:
        abs_array = numpy.log10(abs_array + 1)
        abs_min, abs_max = numpy.log10([abs_min + 1, abs_max + 1])

    normalization = abs_max - abs_min
    normalization = normalization if not numpy.isnan(normalization) else 1.

    norm_abs_array = (abs_array - abs_min) / normalization
    phase_array = (numpy.angle(linearized_complex_array) + numpy.pi) / 2 / numpy.pi

    color_map = cm.get_cmap('hsv')

    colors = color_map(phase_array)
    colors[:, 0] *= norm_abs_array
    colors[:, 1] *= norm_abs_array
    colors[:, 2] *= norm_abs_array

    old_shape = list(complex_array.shape)
    new_shape = old_shape + [4]

    colors = colors.reshape(new_shape)

    return colors


def compute_visibility_range(visibitiles_per_polarization):
    v_max_list = []
    v_min_list = []
    for index, polarization in enumerate(('XX', 'XY', 'YX', 'YY')):
        v_pol = numpy.abs(list(map(lambda x: x[polarization], visibitiles_per_polarization)))
        v_max_list.append(numpy.nanmax(v_pol))
        v_min_list.append(numpy.nanmin(v_pol))

    return min(v_min_list), max(v_max_list)


def _grid_visibilities_lm_plane(l, m, v, sampling):
    dl = 2. / sampling
    over_dl = 1. / dl
    dm = 2. / sampling
    over_dm = 1. / dm

    vis = numpy.zeros((sampling, sampling), dtype=complex)
    l = numpy.array(l)
    m = numpy.array(m)
    l_indexes = numpy.floor((l + 1.) * over_dl).astype(numpy.int)
    m_indexes = numpy.floor((m + 1.) * over_dm).astype(numpy.int)

    for l_i, m_i, v_i in zip(l_indexes, m_indexes, v):
        vis[l_i, m_i] = v_i

    from scipy.ndimage.filters import gaussian_filter

    sigma = [5, 5]
    vis.real = gaussian_filter(vis.real, sigma)
    vis.imag = gaussian_filter(vis.imag, sigma)

    return vis


def _fft_visibilities_lm_plane(l, m, v, sampling):
    vis = _grid_visibilities_lm_plane(l, m, v, sampling)

    fft_vis = fftshift(fft2(vis))
    return fft_vis


def _dft_visibilities_lm_plane(l, m, v, sampling, frequency, array_size):

    x = numpy.linspace(-array_size, array_size, sampling)
    y = numpy.linspace(-array_size, array_size, sampling)
    l = numpy.array(l)
    m = numpy.array(m)
    v = numpy.array(v)
    fft_vis = dft2_numba(x, y, l, m, frequency, v, -1)

    return fft_vis


def _plot_complex_image(canvas, image, extent=None, log=False):
    color_mapped_vis = complex_value_to_color(image, log=log)
    canvas.imshow(color_mapped_vis, extent=extent, origin='lower', resample=True)


def _plot_station_averaged_visibilities_lm_plane_single_frequency(figure: Figure, l_m_v,
                                                                  sampling=512):
    l, m, v, flagged = list(zip(*l_m_v))

    for index, polarization in enumerate(('XX', 'XY', 'YX', 'YY')):
        v_pol = numpy.array(list(map(lambda x: x[polarization], v)))
        canvas = figure.add_subplot(2, 2, index + 1)
        canvas.set_title(polarization)
        canvas.set_xlabel('l')
        canvas.set_ylabel('m')
        canvas.set_xlim(-1, 1)
        canvas.set_ylim(-1, 1)

        vis = _grid_visibilities_lm_plane(l, m, v_pol, sampling)
        _plot_complex_image(canvas, vis, extent=[-1, 1, -1, 1], log=True)


def _plot_station_averaged_visibilities_station_plane_single_frequency(figure: Figure, l_m_v,
                                                                       frequency: float,
                                                                       sampling=50):
    l, m, v, flagged = list(zip(*l_m_v))

    for index, polarization in enumerate(('XX', 'XY', 'YX', 'YY')):
        v_pol = numpy.array(list(map(lambda x: x[polarization], v)))

        array_size = 30
        fft_vis = _dft_visibilities_lm_plane(l=l, m=m, v=v_pol,
                                             sampling=sampling,
                                             frequency=frequency,
                                             array_size=array_size)

        canvas = figure.add_subplot(2, 2, index + 1)
        canvas.set_xlim(-array_size, array_size)
        canvas.set_ylim(-array_size, array_size)

        canvas.set_xlabel('x [m]')
        canvas.set_ylabel('y [m]')

        color_mapped_fft = complex_value_to_color(fft_vis, log=False)
        canvas.imshow(color_mapped_fft[:, ::-1], origin='lower', extent=[-1 * array_size,
                                                                          1 * array_size,
                                                                         -1 * array_size,
                                                                          1 * array_size],
                      resample = True)
        canvas.set_title(polarization)


def _plot_station_averaged_visibilities_lm_plane_datatable(data_table: __DATATABLE_TYPE,
                                                           central_beamlets, save_to=None):

    for station, data_per_station in data_table.items():
        for frequency_str, data_per_frequency in data_per_station.items():

            central_beam = data_per_frequency[central_beamlets[frequency_str]]['mean']

            l_m_v = [(data_per_beam['mean']['l'][0] - central_beam['l'][0],
                      data_per_beam['mean']['m'][0] - central_beam['m'][0],
                      dict(XX=data_per_beam['mean']['XX'][0],
                           XY=data_per_beam['mean']['XY'][0],
                           YX=data_per_beam['mean']['YX'][0],
                           YY=data_per_beam['mean']['YY'][0]),
                      data_per_beam['mean']['flag'][0])
                     for beam, data_per_beam in data_per_frequency.items()]
            frequency = float(frequency_str)

            frequency_in_mhz = frequency / __MHZ_IN_HZ

            figure_title = 'VISIBILITIES_{}-{:4.2f}_MHz.{}'.format(station,
                                                                  frequency_in_mhz,
                                                                  DEFAULT_FORMAT)
            figure = plt.figure(figure_title)


            figure.suptitle(figure_title.replace('_', ' '))

            _plot_station_averaged_visibilities_lm_plane_single_frequency(figure, l_m_v)

            if save_to:
                __save_figure_to_path(figure, save_to, figure_title)
                plt.close(figure)

            figure_title = 'FFT_VISIBILITIES_{}-{:4.2f}_MHz.{}'.format(station,
                                                                       frequency_in_mhz,
                                                                       DEFAULT_FORMAT)
            figure = plt.figure(figure_title)
            figure.suptitle(figure_title.replace('_', ' '))
            _plot_station_averaged_visibilities_station_plane_single_frequency(figure, l_m_v,
                                                                               frequency)

            if save_to:
                __save_figure_to_path(figure, save_to, figure_title)
                plt.close(figure)


def _plot_gains_as_frequency(data_table: __DATATABLE_TYPE, target_station, save_to=None):
    for station, data_per_station in data_table.items():
        opened_figures = set()

        for polarization_index, polarization in enumerate(['XX', 'XY', 'YX', 'YY']):
            extracted_data = [
                (float(frequency_str), data_per_frequency[polarization]['gains'])
                for frequency_str, data_per_frequency in data_per_station.items()]

            frequencies, gains = list(zip(*extracted_data))
            frequencies = numpy.array(frequencies)
            gains = numpy.stack(gains, axis=1)
            for antenna_id, gains_per_antenna in enumerate(gains):
                figure_name = 'gains_{:}_antenna_{:}'.format(target_station,
                                                             antenna_id)
                figure_title = 'GAINS {:} - ANTENNA {:}'.format(target_station, antenna_id)
                figure = plt.figure(figure_name)
                figure.suptitle(figure_title)
                opened_figures.add(figure_name)

                canvas = figure.add_subplot(2, 2, polarization_index + 1)
                canvas.set_title(polarization)
                canvas.set_ylim(-numpy.pi, numpy.pi)

                canvas.plot(frequencies / __MHZ_IN_HZ, numpy.angle(gains_per_antenna), '+-')
        if save_to:
            for figure_name in opened_figures:
                figure = plt.figure(figure_name)
                if save_to:
                    __save_figure_to_path(figure, save_to, figure_name)
                    plt.close(figure)


def __apply_to_datatable(dset: HolographyDataset, derived_datatable_name, function_to_apply, *args,
                         **kwargs):
    if derived_datatable_name in dset.derived_data:
        data_table = dset.derived_data[derived_datatable_name]
        function_to_apply(data_table, *args, **kwargs)


def plot_station_averaged_visibilities_lm_plane(dset: HolographyDataset,
                                                derived_datatable_name='STATION_AVERAGED',
                                                save_to=None):
    __apply_to_datatable(dset,
                         derived_datatable_name,
                         _plot_station_averaged_visibilities_lm_plane_datatable,
                         dset.central_beamlets, save_to=save_to)


def plot_gains_per_antenna(dset: HolographyDataset, derived_datatable_name='GAINS', save_to=None):
    __apply_to_datatable(dset,
                         derived_datatable_name,
                         _plot_gains_as_frequency,
                         dset.target_station_name, save_to=save_to)


def plot_gains_fit_per_antenna(dset: HolographyDataset,
                               interpolation_parameters_datable_name='INTERPOLATED_GAINS',
                               gains_datable_name='GAINS', save_to=None):
    interpolation_parameters = dset.derived_data[interpolation_parameters_datable_name]
    gains = dset.derived_data[gains_datable_name]
    opened_figures = set()

    for station, data_per_station in gains.items():

        for polarization_index, polarization in enumerate(['XX', 'YY']):

            extracted_data = [
                (float(frequency_str), data_per_frequency[polarization]['gains'])
                for frequency_str, data_per_frequency in data_per_station.items()]

            frequencies, gains = list(zip(*extracted_data))
            frequencies = numpy.array(frequencies)
            gains = numpy.stack(gains, axis=1)

            for antenna, antenna_data in enumerate(gains):
                amplitude_par = interpolation_parameters[station][polarization][str(antenna)]['amplitude']['parameters']
                phase_par = \
                interpolation_parameters[station][polarization][str(antenna)]['phase'][
                    'parameters']

                figure_name = 'station_%s_antenna_%s' % (station, antenna)
                figure = plt.figure(figure_name)
                opened_figures.add(figure_name)
                canvas = figure.add_subplot(2, 2, polarization_index + 1)
                __set_minor_ticks_for_canvas(canvas)

                canvas.set_title(polarization)
                canvas.set_xlabel('frequency (MHz)')
                canvas.set_ylabel('Amplitude')
                canvas.plot(frequencies/__MHZ_IN_HZ, numpy.abs(antenna_data), '+')
                canvas.plot(frequencies/__MHZ_IN_HZ, amplitude_par['m'] * frequencies + amplitude_par['q'])

                canvas = figure.add_subplot(2, 2, polarization_index + 3)
                __set_minor_ticks_for_canvas(canvas)

                canvas.set_title(polarization)
                canvas.set_xlabel('frequency (MHz)')
                canvas.set_ylabel('Phase')
                canvas.plot(frequencies/__MHZ_IN_HZ, numpy.angle(antenna_data), '+')
                canvas.plot(frequencies/__MHZ_IN_HZ, phase_par['m'] * frequencies + phase_par['q'])
                plt.tight_layout()
    if save_to:
        for figure_name in opened_figures:
            figure = plt.figure(figure_name)
            if save_to:
                __save_figure_to_path(figure, save_to, figure_name)
                plt.close(figure)
