from typing import Generator, Dict, Tuple, Any
import numpy
import emcee
from scipy.optimize import curve_fit
from scipy.signal import find_peaks

__POSSIBLE_POLARIZATIONS = ('XX', 'YY')

import logging
logger = logging.getLogger(__name__)


def wrap_linear(phase):
    return numpy.angle(numpy.cos(phase) + 1.j * numpy.sin(phase))


def __iterate_gains_over_station_polarization_frequency(dataset: Dict[str,
                                                                        Dict[str,
                                                                           Dict[str,
                                                                             numpy.ndarray]]]) -> \
        Generator[
            Tuple[
                Tuple[str, str],
                Tuple[float],
                numpy.ndarray], None, None]:

    for station_name, dataset_per_station_name in dataset.items():
        sorted_frequencies = sorted(map(float, dataset_per_station_name.keys()))

        for polarization in __POSSIBLE_POLARIZATIONS:
            data_per_polarization = list()

            for frequency in sorted_frequencies:
                data_per_frequency = dataset_per_station_name[str(frequency)][polarization]
                if not data_per_frequency['flag']:

                    data_per_polarization.append(data_per_frequency['gains'])

            data_per_polarization = numpy.stack(data_per_polarization, axis=1)
            yield (station_name, polarization), tuple(sorted_frequencies), data_per_polarization


def __interpolate_gains_per_antenna(dataset: Dict[str, Dict[str, Dict[str, numpy.ndarray]]]):
    data = {}
    for (station, polarization), frequencies, dataset in __iterate_gains_over_station_polarization_frequency(dataset):
        if station not in data:
            data_per_station = {}
            data[station] = data_per_station
        data_per_station = data[station]
        data_per_polarization = dict()
        data_per_station[polarization] = data_per_polarization

        for antenna_id, data_per_antenna in enumerate(dataset):
            logger.info('interpolating gains for polarization %s antenna %s', polarization,
                         antenna_id)
            interpolated_data = _interpolate_gains(frequencies, data_per_antenna)
            data_per_polarization[str(antenna_id)] = interpolated_data
    return data


def _ln_likelihood_abs(parameters, x, y):
    m, q, sigma = parameters
    y_model = m * x + q

    inv_variance = sigma ** -2.
    residual = numpy.square(y - y_model) * inv_variance

    ln_likelihood = -.5 * numpy.sum(residual) - inv_variance * len(y)
    return ln_likelihood


def linear(x, m, q):
    return m * x + q


def _first_guess(x, y, function):
    result, covariance = curve_fit(function, x, y)
    m, q = result
    sigma = numpy.std(x * m + q - y) / numpy.sqrt(len(x) - 3)
    m, q = m * numpy.random.standard_normal(),  q * numpy.random.standard_normal()
    sigma = sigma

    return numpy.array((m, q, sigma))


def __ln_prior(parameters):
    m, b, sigma = parameters
    if sigma >= 0:
        return 0.0
    return -numpy.inf


def __ln_probability(parameters, x, y, likelihood_fn):
    priors = __ln_prior(parameters)
    if numpy.isfinite(priors):
        return priors + likelihood_fn(parameters, x, y)
    return -numpy.inf


def _interpolate_mc_mc(x, y, function, likelihood):

    dims = 3
    n_walkers = 50
    relaxing_steps = 100

    start_position = [_first_guess(x, y, function) for _ in range(n_walkers)]

    sampler = emcee.EnsembleSampler(n_walkers, dims, __ln_probability, args=(x, y, likelihood))
    sampler.run_mcmc(start_position, relaxing_steps, rstate0=1)

    samples = sampler.chain

    m, q, sigma = numpy.median(samples[:, :, 0]),\
                  numpy.median(samples[:, :, 1]),\
                  numpy.median(samples[:, :, 2])

    PSET_TYPES = numpy.dtype([('m', numpy.float32),
                              ('q', numpy.float32),
                              ('sigma', numpy.float32)])

    samples_out = numpy.zeros((n_walkers, relaxing_steps), dtype=PSET_TYPES)
    samples_out['m'] = samples[:, :, 0]
    samples_out['q'] = samples[:, :, 1]
    samples_out['sigma'] = samples[:, :, 2]

    parameters_out = numpy.zeros(1, dtype=PSET_TYPES)
    parameters_out['m'] = m
    parameters_out['q'] = q
    parameters_out['sigma'] = sigma

    return parameters_out, samples_out


def _interpolate_gains(x, y):
    x = numpy.array(x).flatten()
    amplitude = numpy.abs(y).flatten()

    phase = numpy.angle(y).flatten()

    amplitude_parameters, amp_mc_samples = _interpolate_mc_mc(x, amplitude, linear, _ln_likelihood_abs)
    phase_parameters, phase_mc_samples = _interpolate_mc_mc(x, phase, linear, _ln_likelihood_abs)

    result = dict(amplitude=dict(parameters=amplitude_parameters, mc_samples=amp_mc_samples),
                  phase=dict(parameters=phase_parameters, mc_samples=phase_mc_samples))

    return result


def derive_interpolation_parameters(dataset: Dict[str, Dict[str, Dict[str, numpy.ndarray]]]) ->\
        Dict[str, Dict[str, Dict[Any, Dict]]]:
    dataset = __interpolate_gains_per_antenna(dataset)
    return dataset
