import logging
import unittest
from typing import Dict

import lofar.calibration.processing.interpolate as interpolate
import numpy




def generate_signal(nu, slope, intercept, noise):
    signal_amplitude_slope = numpy.abs(slope)
    signal_phase_slope = numpy.angle(slope)

    noise_amplitude = numpy.abs(noise)
    noise_phase = numpy.angle(noise)

    signal = (signal_amplitude_slope * nu +
              abs(intercept) +
              numpy.random.uniform(-noise_amplitude, noise_amplitude, 1)
              ) * numpy.exp((signal_phase_slope * nu + numpy.angle(intercept) +
                             numpy.random.uniform(-noise_phase, noise_phase, 1)) * 1.j)

    return signal


def generate_test_data(slope: complex,
                       intercept: complex,
                       noise_amplitude: complex,
                       n_antennas=40) -> Dict[str, Dict[str, Dict[str, numpy.ndarray]]]:
    """
    As input the routines expects the following structure in the holography dataset
    represented in nested dicts

    [station_name]| all
            + [frequency]
                _+ [polarization]
                    + gains
                    + [...]

    """

    station_set = {'ALL', 'CS001'}
    frequencies = {1., 2., 3., 4.}
    polarizations = ['XX', 'XY', 'YX', 'YY']

    data = lambda nu: numpy.zeros(n_antennas, dtype=numpy.complexfloating) + \
                      generate_signal(nu, slope, intercept, noise_amplitude)
    return {
        station: {
            str(frequency): {
                polarization: {
                    'gains': data(frequency),
                    'flag': numpy.array(False)
                } for polarization in polarizations
            } for frequency in frequencies
        } for station in station_set
    }


class TestInterpolateStep(unittest.TestCase):

    def test_mc_interpolation(self):
        numpy.random.seed(0)
        frequencies = numpy.linspace(0, 10, 15)
        slope = 10 * numpy.exp(.01j)
        intercept = 1. * numpy.exp(+.4j)
        noise_amplitude = 3. * numpy.exp(.005j)

        signal = numpy.array([generate_signal(frequency, slope, intercept, noise_amplitude)
                              for frequency in frequencies])

        result = interpolate._interpolate_gains(frequencies, signal)

        amplitude_slope, amplitude_intercept, amplitude_sigma = result['amplitude']['parameters'][0]
        phase_slope, phase_intercept, phase_sigma = result['phase']['parameters'][0]

        signal_amplitude = numpy.abs(signal)
        signal_phase = numpy.angle(signal)

        expected_residual_amplitude = signal_amplitude - abs(slope) * frequencies - abs(intercept)
        expected_residual_phase = signal_amplitude - numpy.angle(slope) * frequencies - numpy.angle(intercept)

        residual_amplitude = signal_amplitude - amplitude_slope * frequencies - amplitude_intercept
        residual_phase = signal_phase - phase_slope * frequencies - phase_intercept

        error_amplitude = 1 - numpy.std(residual_amplitude) / numpy.std(expected_residual_amplitude)
        error_phase = numpy.std(residual_phase) / numpy.std(expected_residual_phase)

        self.assertLessEqual(error_amplitude, .05)
        self.assertLessEqual(error_phase, .05)

    def test_interpolation(self):
        slope = 10 * numpy.exp(1.j)
        intercept = 1. * numpy.exp(-.4j)
        noise_amplitude = 1  # 3.*numpy.exp(.001j)
        test_data = generate_test_data(slope, intercept, noise_amplitude, n_antennas=4)
        result = interpolate.derive_interpolation_parameters(test_data)
        self.assertEqual(test_data.keys(), result.keys())


if __name__ == '__main__':
    logging.basicConfig(format='%(name)s : %(message)s')
    unittest.main()
